import { AutocompleteInteraction, PermissionsBitField } from "discord.js";
import { RESTPostAPIApplicationCommandsJSONBody } from "discord.js";
import { APIApplicationCommandOption, ApplicationCommandType, ChatInputCommandInteraction, LocalizationMap, MessageInteraction, PermissionResolvable, UserSelectMenuInteraction } from "discord.js";

export type InteractionTypeMap = {
  // [CommandType]: [Interaction, Description]
  [ApplicationCommandType.ChatInput]: [ChatInputCommandInteraction, string];
  [ApplicationCommandType.Message]: [MessageInteraction, never];
  [ApplicationCommandType.User]: [UserSelectMenuInteraction, never];
};

// TODO: At time of coding, Discord api types doesn't support user installations of bot/application yet
//       replace this with the types from the discord api types when it's available
/**
 * https://discord.com/developers/docs/resources/application#application-object-application-integration-types
 */
export enum ApplicationIntegrationType {
  Guild_Install = 0,
  User_Install = 1,
}

/**
 * https://discord.com/developers/docs/interactions/receiving-and-responding#interaction-object-interaction-context-types
 */
export enum InteractionContextTypes {
  Guild = 0,
  BotDM = 1,
  PrivateChannel = 2,
}

/**
 * https://discord.com/developers/docs/interactions/application-commands#application-command-object-application-command-structure
 */
export type FutureRESTPostAPIApplicationCommandsJSONBody =
  RESTPostAPIApplicationCommandsJSONBody
& {
  /**
   * @deprecated use contexts instead
   */
  dm_permission?: boolean;
  integration_types?: ApplicationIntegrationType[];
  contexts?: InteractionContextTypes[];
};


interface Command<Type extends keyof InteractionTypeMap = ApplicationCommandType> {
  readonly name: string;
  readonly name_localizations?: LocalizationMap;
  readonly description: InteractionTypeMap[Type][1];
  readonly description_localizations?: LocalizationMap;
  readonly options?: APIApplicationCommandOption[];
  readonly default_member_permissions?: PermissionResolvable;
  readonly type: Type;
  readonly nsfw?: boolean;
  /** @deprecated use contexts instead */
  readonly dm_permission?: boolean;
  readonly integration_types?: ApplicationIntegrationType[];
  readonly contexts?: InteractionContextTypes[];
}

abstract class Command<Type extends keyof InteractionTypeMap = ApplicationCommandType> {
  abstract execute(interaction: InteractionTypeMap[Type][0]): Promise<void>;

  autocomplete?(interaction: Type extends ApplicationCommandType.ChatInput ? AutocompleteInteraction : never ): Promise<void>;

  toRESTPostApplicationCommands(): FutureRESTPostAPIApplicationCommandsJSONBody {
    return {
      name: this.name,
      name_localizations: this.name_localizations,
      description: this.description,
      description_localizations: this.description_localizations,
      options: this.options,
      default_member_permissions: this.default_member_permissions !== undefined ? new PermissionsBitField(this.default_member_permissions).bitfield.toString() : undefined,
      type: this.type,
      nsfw: this.nsfw,
      dm_permission: this.dm_permission,
      integration_types: this.integration_types,
      contexts: this.contexts,
    };
  }
}

export default Command;
